package
{
	import alternativa.engine3d.core.Object3D;
	import alternativa.engine3d.core.events.MouseEvent3D;
	import alternativa.engine3d.lights.OmniLight;
	import alternativa.engine3d.materials.TextureMaterial;
	import alternativa.engine3d.materials.VertexLightTextureMaterial;
	import alternativa.engine3d.primitives.Plane;
	import alternativa.engine3d.resources.BitmapTextureResource;
	
	import assets.BgAsset;
	import assets.GradationAsset;
	import assets.PreloaderAsset;
	
	import flash.display.BitmapData;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.events.TimerEvent;
	import flash.geom.Point;
	import flash.net.URLRequest;
	import flash.ui.Keyboard;
	import flash.utils.Timer;
	
	import jp.clockmaker.prog4.commands.LimitationParallelList;
	import jp.nium.utils.NumberUtil;
	import jp.progression.commands.Func;
	import jp.progression.commands.Wait;
	import jp.progression.commands.lists.SerialList;
	import jp.progression.commands.net.LoadBitmapData;
	import jp.progression.data.getResourceById;
	import jp.progression.events.ExecuteEvent;
	
	import org.libspark.alternativa3d.view.AlternativaTemplate;
	import org.libspark.betweenas3.BetweenAS3;
	import org.libspark.betweenas3.easing.Expo;
	import org.libspark.betweenas3.tweens.ITween;
	
	import project.ui.ScrollBar;

	[SWF(frameRate = "60", backgroundColor = "#000000")]
	/**
	 * Cover Flow サンプル
	 * @author yasu
	 */
	public class Main extends AlternativaTemplate
	{

		//----------------------------------------------------------
		//
		//   Constructor 
		//
		//----------------------------------------------------------

		public function Main()
		{
			super({
				antiAlias:2,
				useDiagram:true
			});
		}

		//----------------------------------------------------------
		//
		//   Property 
		//
		//----------------------------------------------------------

		private const ITEM_H:Number = 256; // 平面の縦幅
		private const ITEM_W:Number = 256; // 平面の横幅
		private const MARGIN_X:Number = 80; // 平面のX座標の間隔
		private const MAX_SLIDE:int = 200; // ページの個数
		private var currentPage:int = 0; // 現在のページID
		private var gradationBmd:BitmapData;
		private var materialsOrijinal:Array /*index*/;
		private var materialsReflection:Array /*index*/;
		private var planeArr:Vector.<Object3D> = new <Object3D>[]; // 平面を格納する配列
		private var preloader:PreloaderAsset;
		private var resizeTimer:Timer;
		private var scrollBar:ScrollBar; // スクロールバー
		private var tween:ITween;
		private var urls:Vector.<String> = new <String>[];

		//----------------------------------------------------------
		//
		//   Function 
		//
		//----------------------------------------------------------

		override protected function atInit():void
		{
			// カメラの座標
			cameraController.setObjectPosXYZ(0, -510, 0);
			camera.y = -510;
			cameraController.update();

			// 処理のための準備
			// グラデーションマスク
			gradationBmd = new BitmapData(256, 256, true, 0x00000000);
			gradationBmd.draw(new GradationAsset());

			// プリローダー
			preloader = new PreloaderAsset();
			preloader.label_txt.text = "0/100";
			addChild(preloader);

			// マテリアルの格納場所(仕方なくArrayで)
			materialsOrijinal = [];
			materialsReflection = [];
			
			var list:LimitationParallelList = new LimitationParallelList(4);
			for (var i:int = 0; i < MAX_SLIDE; i++)
			{
				urls[i] = "imgs/" + NumberUtil.digit(i%40, 2) + ".jpg";

				list.addCommand(
					new SerialList(null,
					new LoadBitmapData(new URLRequest(urls[i])),
					new Func(renderTexture, [i]),
					new Wait(0.016)
					)
					);
			}
			list.onPosition = function():void {
				preloader.label_txt.text = int(list.completePosition / list.numCommands * 100) + "/100";
			};
			list.addEventListener(ExecuteEvent.EXECUTE_COMPLETE, list_executeCompleteHandler);
			list.execute();

			// リサイズ関係
			resizeTimer = new Timer(100, 1);
			resizeTimer.addEventListener(TimerEvent.TIMER_COMPLETE, resizeTimer_timerCompleteHandler);
			onResize();
		}

		/**
		 * 読み込み完了後
		 */
		private function list_executeCompleteHandler(event:ExecuteEvent):void
		{
			preloader.label_txt.text = "100/100";
			
			new SerialList(null,
				new Wait(0.05),
				new Func(removeChild, [preloader]),
				new Func(initialize3D)
			).execute();
		}

		override protected function onResize(e:Event = null):void
		{
			if (scrollBar)
			{
				scrollBar.x = (stage.stageWidth - scrollBar.width) / 2 >> 0;
				scrollBar.y = stage.stageHeight - 70;
			}

			if (preloader)
			{
				preloader.x = stage.stageWidth / 2 >> 0;
				preloader.y = stage.stageHeight / 2 >> 0;
			}

			super.onResize(e);

			if (resizeTimer)
			{
				resizeTimer.reset();
				resizeTimer.start();
			}
		}

		override protected function onRenderTick(e:Event = null):void
		{
			if (tween && tween.isPlaying)
				super.onRenderTick(e);
		}

		private function renderTexture(index:int):void
		{
			var zero:Point = new Point();
			var bmd:BitmapData = getResourceById(urls[index]).toBitmapData();

			var bmdOpt:BitmapData = new BitmapData(256, 256, true, 0x0);
			bmdOpt.copyPixels(bmd, bmd.rect, zero, gradationBmd, zero, true);
			var texture:BitmapTextureResource = new BitmapTextureResource(bmd);
			var textureOpt:BitmapTextureResource = new BitmapTextureResource(bmdOpt);

			// マテリアルの作成
			var material:VertexLightTextureMaterial = new VertexLightTextureMaterial(texture);
			var materialOpt:VertexLightTextureMaterial = new VertexLightTextureMaterial(textureOpt, null, 0.3);

			materialsOrijinal[index] = material;
			materialsReflection[index] = materialOpt;
		}

		private function resizeTimer_timerCompleteHandler(event:TimerEvent):void
		{
			// レンダリング
			camera.render(stage3D);
		}

		private function initialize3D():void
		{
			// 照明 (遠近感をつけるために・・・)
			var light:OmniLight = new OmniLight(0xFFFFFF, 1, 1000);
			light.y = -ITEM_W;
			scene.addChild(light);
			var light2:OmniLight = new OmniLight(0xFFFFFF, 1, 1000);
			light2.y = -ITEM_W;
			scene.addChild(light2);
			
			// 背景
			var bgBmd:BitmapData = new BitmapData(256, 256);
			bgBmd.draw(new BgAsset());
			var bgPlane:Plane = new Plane(2200, 1200, 1, 1, false, false, null, new TextureMaterial(new BitmapTextureResource(bgBmd)));
			bgPlane.rotationX = Math.PI / 2;
			bgPlane.y = 500;
			scene.addChild(bgPlane);

			// Planeの作成
			for (var i:int = 0; i < MAX_SLIDE; i++)
			{
				// 平面の作成
				var plane:Object3D = new Object3D();
				plane.addEventListener(MouseEvent3D.CLICK, plane_clickHandler);

				// 上面
				var planeTop:Plane = new Plane(ITEM_W, ITEM_H, 1, 1, false, false, null, materialsOrijinal[i]);
				planeTop.rotationX = Math.PI / 2;
				plane.addChild(planeTop);

				// 反射面
				var planeBottom:Plane = new Plane(ITEM_W, ITEM_H, 1, 1, false, true, null, materialsReflection[i]);
				planeBottom.rotationX = -Math.PI / 2;
				planeBottom.z = -ITEM_H - 2;
				plane.addChild(planeBottom);

				// 3Dシーンに追加
				scene.addChild(plane);

				// 配列に参照の保存
				planeArr[i] = plane;
			}
			planeArr.fixed = true;

			// テクスチャのアップロード
			uploadResource();

			// 初期のページ表示
			moveSlide(MAX_SLIDE / 2);

			// スクロールバーの作成
			scrollBar = new ScrollBar();
			scrollBar.value = 0.5;
			scrollBar.addEventListener(Event.CHANGE, scrollbar_changeHandler);
			addChild(scrollBar);

			// UIの制御
			stage.addEventListener(KeyboardEvent.KEY_DOWN, stage_keyDownHandler);
			stage.addEventListener(MouseEvent.MOUSE_WHEEL, stage_mouseWheelHandler);

			// リサイズ
			onResize();
		}

		private function stage_mouseWheelHandler(event:MouseEvent):void
		{
			// スクロールバーの更新
			scrollBar.value += event.delta / 4 * 1 / (MAX_SLIDE - 1);
			scrollbar_changeHandler(null);
		}

		private function stage_keyDownHandler(event:KeyboardEvent):void
		{
			var direction:int = 0;
			switch (event.keyCode)
			{
				case Keyboard.LEFT:
				case Keyboard.DOWN:
				{
					direction = -1;
					break;
				}
				case Keyboard.RIGHT:
				case Keyboard.UP:
				{
					direction = 1;
					break;
				}
			}

			if (direction != 0)
			{
				// スクロールバーの更新
				scrollBar.value += direction * 1 / (MAX_SLIDE - 1);
				scrollbar_changeHandler(null);
			}
		}

		/**
		 * スライドを移動
		 * @param id	スライドのID
		 */
		private function moveSlide(id:uint):void
		{
			// 遷移先が現在のスライド番号と同じであれば処理を終了
			if (currentPage == id)
				return;

			var array:Array = [];

			for (var i:uint = 0; i < MAX_SLIDE; i++)
			{
				// 移動値を初期化
				var targetX:Number = 0;
				var targetY:Number = 0;
				var targetRot:Number = 0;

				// X座標の計算
				targetX = MARGIN_X * (i - id);

				// 中央のスライド画像より左側のもの
				if (i < id)
				{
					targetX -= ITEM_W * 0.6;
					targetY = ITEM_W;
					targetRot = 45 * Math.PI / 180;
				}
				// 中央のスライド画像より右側のもの
				else if (i > id)
				{
					targetX += ITEM_W * 0.6;
					targetY = ITEM_W;
					targetRot = -45 * Math.PI / 180;
				}
				// 中央のスライド画像
				else
				{
					targetX += 0;
					targetY = 0;
					targetRot = 0;
				}

				// 対象のPlaneの参照をpに格納
				var p:Object3D = planeArr[i];

				// 0.9秒で角度をアニメーション
				array.push(BetweenAS3.to(p, {
						rotationZ: targetRot
					}, 0.9, Expo.easeOut));

				// 1.8秒で位置をアニメーション
				array.push(BetweenAS3.to(p, {
						x: targetX,
						y: targetY
					}, 1.8, Expo.easeOut));
			}

			currentPage = id;

			// 不要なトゥイーンは停止
			if (tween && tween.isPlaying)
				tween.stop();

			// 全部まとめてトゥイーン
			tween = BetweenAS3.parallelTweens(array);
			tween.play();
		}

		/**
		 * Planeをクリックしたときのイベント
		 */
		private function plane_clickHandler(e:MouseEvent3D):void
		{
			// Planeのextraプロパティに格納したページIDの取得
			var nextId:int = planeArr.indexOf(e.currentTarget);

			if (nextId > -1)
			{
				// ページ遷移
				moveSlide(nextId);
				// スクロールバーの更新
				scrollBar.value = nextId / (MAX_SLIDE - 1);
			}
		}

		/**
		 * スクロールが動いたときのイベント
		 */
		private function scrollbar_changeHandler(e:Event):void
		{
			// スクロールバーの値からページIDの計算
			var nextId:int = Math.round(scrollBar.value * (MAX_SLIDE - 1));
			// ページ遷移
			moveSlide(nextId);
		};
	}
}
