package jp.clockmaker.prog4.commands {
	import jp.nium.core.debug.Logger;
	import jp.progression.commands.Command;
	import jp.progression.commands.CommandList;
	import jp.progression.commands.Func;
	import jp.progression.commands.Trace;
	import jp.progression.commands.Wait;
	import jp.progression.commands.lists.ParallelList;
	import jp.progression.commands.lists.SerialList;
	import jp.progression.core.L10N.L10NCommandMsg;
	import jp.progression.events.ExecuteErrorEvent;
	import jp.progression.events.ExecuteEvent;

	/**
	 * Progression 4 用の上限付き並列実行コマンド
	 * @version 0.1
	 * @update 2009/11/04
	 */
	public class LimitationParallelList extends CommandList {
		/**
		 * 現在処理中の Command インスタンスを保存した配列を取得します。
		 */
		private var _commands:Array;
		/**
		 * 同時接続数を取得します。
		 */
		private var _maxConnectionCount:int;

		public function get maxConnectionCount():int {
			return _maxConnectionCount;
		}
		/**
		 * 実行中のコマンド数を取得します。
		 */
		private var _executeCount:int;
		/**
		 * 実行済みのコマンド数を取得します。
		 */
		private var _completePosition:int;

		public function get completePosition():int {
			return _completePosition;
		}

		/**
		 * <p>コマンドオブジェクトが ExecuteEvent.EXECUTE_POSITION イベントを受け取った場合に呼び出されるイベントハンドラメソッドを取得または設定します。</p>
		 * <p></p>
		 *
		 * @see #scope
		 * @see jp.progression.events.ExecuteEvent#POSITION
		 */
		public function get onPosition():Function {
			return _onPosition;
		}

		public function set onPosition(value:Function):void {
			_onPosition = value;
		}
		private var _onPosition:Function;

		/**
		 * <p>新しい MyLoaderList インスタンスを作成します。</p>
		 * <p>Creates a new MyLoaderList object.</p>
		 *
		 * @param initObject
		 * <p>設定したいプロパティを含んだオブジェクトです。</p>
		 * <p></p>
		 * @param commands
		 * <p>登録したいコマンドインスタンス、関数、数値、配列などを含む配列です。</p>
		 * <p></p>
		 */
		public function LimitationParallelList(maxConnectionCount:int = 6, initObject:Object = null, ... commands:Array) {
			// 初期化する
			_maxConnectionCount = maxConnectionCount;
			_commands = [];
			_executeCount = 0;
			// 親クラスを初期化する
			super(executeFunction, interruptFunction, initObject);
			// initObject に自身のサブクラス以外のコマンドが指定されたら警告する
			var com:Command = initObject as Command;
			if (com && !(com is LimitationParallelList)) {
			}
			// コマンドリストに登録する
			addCommand.apply(null, commands);
		}

		/**
		 * @private
		 */
		protected function executeFunction():void {
			// 初期化する
			super.reset();
			_commands = [];
			_completePosition = 0;
			// 現在登録されている全てのコマンドを取得する
			while (super.hasNextCommand()) {
				var com:Command = super.nextCommand();
				// イベントリスナーを登録する
				com.addEventListener(ExecuteEvent.EXECUTE_POSITION, dispatchEvent);
				com.addEventListener(ExecuteEvent.EXECUTE_COMPLETE, _executeComplete);
				com.addEventListener(ExecuteEvent.EXECUTE_INTERRUPT, _executeInterrupt);
				com.addEventListener(ExecuteErrorEvent.EXECUTE_ERROR, _error);
				// 登録する
				_commands.push(com);
			}
			// 登録されていれば
			if (_commands.length) {
				// コマンドを同時に実行する
				var commands:Array = _commands.slice();
				for (var i:int = 0, l:int = commands.length; i < l; i++) {
					com = commands[i];
					// イベントを送出する
					super.dispatchEvent(new ExecuteEvent(ExecuteEvent.EXECUTE_POSITION, false, false, com));
					// 実行中であれば実行する
					if (state == 2) {
						// 上限数を超えていれば待機する
						if (_maxConnectionCount > _executeCount) {
							com.execute(extra);
							_executeCount++;
						}
					}
				}
				import flash.trace.Trace;

				import jp.progression.commands.Func;
				import jp.progression.commands.Trace;
				import jp.progression.commands.Wait;
				import jp.progression.commands.lists.ParallelList;
				import jp.progression.commands.lists.SerialList;

			} else {
				// 処理を終了する
				super.executeComplete();
			}
		}

		/**
		 * @private
		 */
		protected function interruptFunction():void {
			// コマンドを同時に中断する
			var commands:Array = _commands.slice();
			for (var i:int = 0, l:int = commands.length; i < l; i++) {
				var com:Command = commands[i];
				// イベントリスナーを解除する
				com.removeEventListener(ExecuteEvent.EXECUTE_POSITION, dispatchEvent);
				com.removeEventListener(ExecuteEvent.EXECUTE_COMPLETE, _executeComplete);
				com.removeEventListener(ExecuteEvent.EXECUTE_INTERRUPT, _executeInterrupt);
				com.removeEventListener(ExecuteErrorEvent.EXECUTE_ERROR, _error);
				// 実行中であれば中断する
				if (com.state > 0) {
					com.interrupt();
				}
			}
			// 初期化する
			_commands = [];
		}

		/**
		 * 破棄します。
		 */
		private function _destroy(command:Command):void {
			if (command) {
				// イベントリスナーを解除する
				command.removeEventListener(ExecuteEvent.EXECUTE_POSITION, dispatchEvent);
				command.removeEventListener(ExecuteEvent.EXECUTE_COMPLETE, _executeComplete);
				command.removeEventListener(ExecuteEvent.EXECUTE_INTERRUPT, _executeInterrupt);
				command.removeEventListener(ExecuteErrorEvent.EXECUTE_ERROR, _error);
				_completePosition++;
				// イベントハンドラメソッドを実行する
				if (_onPosition != null) {
					_onPosition.apply(scope || this);
				}
			}
		}

		/**
		 * <p>登録されているコマンドの最後尾に新しくコマンドを追加します。
		 * 関数を指定した場合には、自動的に Func コマンドに変換され、
		 * 数値を指定した場合には、自動的に Wait コマンドに変換され、
		 * 配列を指定した場合には、自動的に SerialList コマンドに変換され、
		 * それ以外の値は、全て自動的に Trace コマンドに変換されます。</p>
		 * <p></p>
		 *
		 * @see #position
		 * @see #insertCommand()
		 * @see #clearCommand()
		 *
		 * @param commands
		 * <p>登録したいコマンドを含む配列です。</p>
		 * <p></p>
		 */
		public override function addCommand(... commands:Array):void {
			// 親のメソッドを実行する
			//			super.addCommand.apply(null, CommandUtil.convert(this, commands));
			super.addCommand.apply(null, _convert(this, commands));
		}

		/**
		 * <p>現在実行中のコマンドの次の位置に新しくコマンドを追加します。
		 * 関数を指定した場合には、自動的に Func コマンドに変換され、
		 * 数値を指定した場合には、自動的に Wait コマンドに変換され、
		 * 配列を指定した場合には、自動的に SerialList コマンドに変換され、
		 * それ以外の値は、全て自動的に Trace コマンドに変換されます。</p>
		 * <p></p>
		 *
		 * @see #position
		 * @see #addCommand()
		 * @see #clearCommand()
		 *
		 * @param commands
		 * <p>登録したいコマンドを含む配列です。</p>
		 * <p></p>
		 * @return
		 * <p>自身の参照です。</p>
		 * <p></p>
		 */
		public override function insertCommand(... commands:Array):void {
			// 親のメソッドを実行する
			//			super.insertCommand.apply(null, CommandUtil.convert(this, commands));
			super.addCommand.apply(null, _convert(this, commands));
		}

		/**
		 * <p>MyLoaderList インスタンスのコピーを作成して、各プロパティの値を元のプロパティの値と一致するように設定します。</p>
		 * <p>Duplicates an instance of an MyLoaderList subclass.</p>
		 *
		 * @return
		 * <p>元のオブジェクトと同じプロパティ値を含む新しい MyLoaderList インスタンスです。</p>
		 * <p>A new MyLoaderList object that is identical to the original.</p>
		 */
		public override function clone():Command {
			return new LimitationParallelList(_maxConnectionCount, this);
		}

		/**
		 * コマンドの処理が完了した場合に送出されます。
		 */
		private function _executeComplete(e:ExecuteEvent):void {
			// 破棄する
			_destroy(Command(e.target));
			// まだ存在すれば終了する
			if (_commands.length > _completePosition) {
				// 次の処理を開始する
				if (_commands.length > _executeCount) {
					// イベントを送出する
					var com:Command = _commands[_executeCount];
					super.dispatchEvent(new ExecuteEvent(ExecuteEvent.EXECUTE_POSITION, false, false, com));
					com.execute(extra);
					_executeCount++;
				}
				return;
			}
			// 実行中であれば終了する
			if (super.state > 0) {
				super.executeComplete();
			}
		}

		/**
		 * コマンドの処理が中断された場合に送出されます。
		 */
		private function _executeInterrupt(e:ExecuteEvent):void {
			if (e.enforcedInterrupting) {
				// 処理を中断する
				super.interrupt(e.enforcedInterrupting);
			} else {
				// 処理を完了する
				super.executeComplete();
			}
		}

		/**
		 * 処理の途中でエラーが発生した場合に送出されます。
		 */
		private function _error(e:ExecuteErrorEvent):void {
			// エラー処理を実行する
			super.throwError(e.errorTarget as Command, e.errorObject);
		}


		/**
		 * @private
		 */
		internal static function _convert(target:Command, commands:Array):Array {
			for (var i:int = 0, l:int = commands.length; i < l; i++) {
				var value:* = commands[i];

				// 型に応じて変換する
				switch (true) {
					// Command インスタンス、または null であれば次へ
					case value == undefined:
					case value == null:
					case value is Command:  {
						break;
					}

					// 関数であれば Func コマンドに変換する
					case value is Function:  {
						commands[i] = new Func(value as Function);
						break;
					}

					// 数値であればディレイ付き Command コマンドに変換する
					case value is Number:  {
						commands[i] = new Wait(value);
						break;
					}

					// 配列であればシンタックスシュガーとして処理する
					case value is Array:  {
						var list:CommandList;

						switch (true) {
							case target is SerialList:  {
								list = new ParallelList();
								break;
							}
							case target is ParallelList:
							default:  {
								list = new SerialList();
							}
						}

						list.addCommand.apply(null, value as Array);
						commands[i] = list;
						break;
					}

					// それ以外は全て Trace コマンドに変換する
					default:  {
						commands[i] = new Trace("trace: " + value);
					}
				}
			}

			return commands;
		}


	}
}